@@ -1,35 +1,38 @@
 #Auxiliary functions; some from WeightIt
 
 #Function to process inputs and throw warnings or errors if inputs are incompatible with methods
-check.inputs <- function(method, distance, mcall, exact, mahvars, caliper, discard, reestimate, s.weights, replace, ratio, m.order, estimand) {
+check.inputs <- function(mcall, method, distance, exact, mahvars, antiexact,
+                         caliper, discard, reestimate, s.weights, replace,
+                         ratio, m.order, estimand, ...,
+                         min.controls = NULL, max.controls = NULL) {
 
   null.method <- is.null(method)
   if (null.method) {
     method <- "NULL"
   }
   else {
-    method <- match_arg(method, c("exact", "cem", "nearest", "optimal", "full", "genetic", "subclass"))
+    method <- match_arg(method, c("exact", "cem", "nearest", "optimal", "full", "genetic", "subclass", "cardinality"))
   }
 
   ignored.inputs <- character(0)
   error.inputs <- character(0)
   if (null.method) {
-    for (i in c("exact", "mahvars", "caliper", "std.caliper", "replace", "ratio", "m.order")) {
-      if (i %in% names(mcall) && !is.null(get0(i))) {
+    for (i in c("exact", "mahvars", "antiexact", "caliper", "std.caliper", "replace", "ratio", "min.controls", "max.controls", "m.order")) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
         ignored.inputs <- c(ignored.inputs, i)
       }
     }
   }
   else if (method == "exact") {
-    for (i in c("distance", "exact", "mahvars", "caliper", "std.caliper", "discard", "reestimate", "replace", "ratio", "m.order")) {
-      if (i %in% names(mcall) && !is.null(get0(i))) {
+    for (i in c("distance", "exact", "mahvars", "antiexact", "caliper", "std.caliper", "discard", "reestimate", "replace", "ratio", "min.controls", "max.controls", "m.order")) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
         ignored.inputs <- c(ignored.inputs, i)
       }
     }
   }
   else if (method == "cem") {
-    for (i in c("distance", "exact", "mahvars", "caliper", "std.caliper", "discard", "reestimate", "replace", "ratio", "m.order")) {
-      if (i %in% names(mcall) && !is.null(get0(i))) {
+    for (i in c("distance", "exact", "mahvars", "antiexact", "caliper", "std.caliper", "discard", "reestimate", "replace", "ratio", "min.controls", "max.controls", "m.order")) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
         ignored.inputs <- c(ignored.inputs, i)
       }
     }
@@ -37,7 +40,7 @@ check.inputs <- function(method, distance, mcall, exact, mahvars, caliper, disca
   else if (method == "nearest") {
     if (is.character(distance) && distance == "mahalanobis") {
       for (e in c("mahvars", "reestimate")) {
-        if (e %in% names(mcall) && !is.null(get0(e))) {
+        if (e %in% names(mcall) && !is.null(e_ <- get0(e, inherits = FALSE)) && !identical(e_, formals(MatchIt::matchit)[[e]])) {
           error.inputs <- c(error.inputs, e)
         }
       }
@@ -46,14 +49,14 @@ check.inputs <- function(method, distance, mcall, exact, mahvars, caliper, disca
   else if (method == "optimal") {
     if (is.character(distance) && distance == "mahalanobis") {
       for (e in c("mahvars", "reestimate")) {
-        if (e %in% names(mcall) && !is.null(get0(e))) {
+        if (e %in% names(mcall) && !is.null(e_ <- get0(e, inherits = FALSE)) && !identical(e_, formals(MatchIt::matchit)[[e]])) {
           error.inputs <- c(error.inputs, e)
         }
       }
     }
 
     for (i in c("replace", "caliper", "std.caliper", "m.order")) {
-      if (i %in% names(mcall) && !is.null(get0(i))) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
         ignored.inputs <- c(ignored.inputs, i)
       }
     }
@@ -62,14 +65,14 @@ check.inputs <- function(method, distance, mcall, exact, mahvars, caliper, disca
   else if (method == "full") {
     if (is.character(distance) && distance == "mahalanobis") {
       for (e in c("mahvars", "reestimate")) {
-        if (e %in% names(mcall) && !is.null(get0(e))) {
+        if (e %in% names(mcall) && !is.null(e_ <- get0(e, inherits = FALSE)) && !identical(e_, formals(MatchIt::matchit)[[e]])) {
           error.inputs <- c(error.inputs, e)
         }
       }
     }
 
     for (i in c("replace", "ratio", "m.order")) {
-      if (i %in% names(mcall) && !is.null(get0(i))) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
         ignored.inputs <- c(ignored.inputs, i)
       }
     }
@@ -77,19 +80,31 @@ check.inputs <- function(method, distance, mcall, exact, mahvars, caliper, disca
   else if (method == "genetic") {
     if (is.character(distance) && distance == "mahalanobis") {
       for (e in c("mahvars", "reestimate")) {
-        if (e %in% names(mcall) && !is.null(get0(e))) {
+        if (e %in% names(mcall) && !is.null(e_ <- get0(e, inherits = FALSE)) && !identical(e_, formals(MatchIt::matchit)[[e]])) {
           error.inputs <- c(error.inputs, e)
         }
       }
     }
+    for (i in c("min.controls", "max.controls")) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
+        ignored.inputs <- c(ignored.inputs, i)
+      }
+    }
+  }
+  else if (method == "cardinality") {
+    for (i in c("distance", "mahvars", "antiexact", "caliper", "std.caliper", "reestimate", "replace", "min.controls", "m.order")) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
+        ignored.inputs <- c(ignored.inputs, i)
+      }
+    }
   }
   else if (method == "subclass") {
     if (is.character(distance) && distance == "mahalanobis") {
       stop("distance = \"mahalanobis\" is not compatible with subclassification.", call. = FALSE)
     }
 
-    for (i in c("exact", "mahvars", "caliper", "std.caliper", "replace", "ratio", "m.order")) {
-      if (i %in% names(mcall) && !is.null(get0(i))) {
+    for (i in c("exact", "mahvars", "antiexact", "caliper", "std.caliper", "replace", "ratio", "min.controls", "max.controls", "m.order")) {
+      if (i %in% names(mcall) && !is.null(i_ <- get0(i, inherits = FALSE)) && !identical(i_, formals(MatchIt::matchit)[[i]])) {
         ignored.inputs <- c(ignored.inputs, i)
       }
     }
@@ -105,24 +120,26 @@ check.inputs <- function(method, distance, mcall, exact, mahvars, caliper, disca
                                             " not allowed with method = ", add_quotes(method, quotes = !null.method),
                                             " and distance = \"", distance, "\"."),
                                      call. = FALSE)
+  return(ignored.inputs)
 }
 
 #Function to process distance and give warnings about new syntax
-process.distance <- function(distance, method) {
+process.distance <- function(distance, method = NULL, treat) {
   if (is.null(distance) && !is.null(method)) stop(paste0("'distance' cannot be NULL with method = \"", method, "\"."), call. = FALSE)
-  else if (is.vector(distance, "character") && length(distance) == 1) {
-    allowable.distances <- c("glm", "cbps", "gam", "mahalanobis", "nnet", "rpart", "bart", "randomforest")
+  else if (is.character(distance) && length(distance) == 1) {
+    allowable.distances <- c("glm", "cbps", "gam", "mahalanobis", "nnet", "rpart", "bart",
+                             "randomforest", "elasticnet", "lasso", "ridge", "gbm")
 
     if (tolower(distance) %in% c("cauchit", "cloglog", "linear.cloglog", "linear.log", "linear.logit", "linear.probit",
-                        "linear.cauchit", "log", "probit")) {
+                                 "linear.cauchit", "log", "probit")) {
       warning(paste0("'distance = \"", distance, "\"' will be deprecated; please use 'distance = \"glm\", link = \"", distance, "\"' in the future."), call. = FALSE, immediate. = TRUE)
       link <- distance
       distance <- "glm"
       attr(distance, "link") <- link
     }
     else if (tolower(distance) %in% tolower(c("GAMcloglog", "GAMlog", "GAMlogit", "GAMprobit"))) {
-      warning(paste0("'distance = \"", distance, "\"' will be deprecated; please use 'distance = \"gam\", link = \"", sub("GAM", "", distance), "\"' in the future."), call. = FALSE, immediate. = TRUE)
-      link <- distance
+      link <- sub("GAM", "", distance)
+      warning(paste0("'distance = \"", distance, "\"' will be deprecated; please use 'distance = \"gam\", link = \"", link, "\"' in the future."), call. = FALSE, immediate. = TRUE)
       distance <- "gam"
       attr(distance, "link") <- link
     }
@@ -130,48 +147,130 @@ process.distance <- function(distance, method) {
       distance <- "glm"
       attr(distance, "link") <- "logit"
     }
+    else if (tolower(distance) == "glmnet") {
+      distance <- "elasticnet"
+    }
     else if (!tolower(distance) %in% allowable.distances) {
-      stop("The argument supplied to distance is not an allowable value. See ?matchit for allowable options.", call. = FALSE)
+      stop("The argument supplied to distance is not an allowable value. See ?distance for allowable options.", call. = FALSE)
     }
     else {
       distance <- tolower(distance)
     }
 
   }
-  else if (!is.vector(distance, "numeric")) {
-    stop("'distance' must be a string with the name of the distance measure to be used or a numeric vector containing distance measures.", call. = FALSE)
+  else if (!is.numeric(distance) || (!is.null(dim(distance)) && length(dim(distance)) != 2)) {
+    stop("'distance' must be a string with the name of the distance measure to be used or a numeric vector or matrix containing distance measures.", call. = FALSE)
+  }
+  else if (is.matrix(distance) && (is.null(method) || !method %in% c("nearest", "optimal", "full"))) {
+    if (is.null(method)) method <- "NULL" else method <- paste0('"', method, '"')
+    stop(paste0("'distance' cannot be supplied as a matrix with method = ", method, "."), call. = FALSE)
+  }
+
+  if (is.numeric(distance)) {
+    if (is.matrix(distance)) {
+      dim.distance <- dim(distance)
+      if (all(dim.distance == length(treat))) {
+        if (!is.null(rownames(distance))) distance <- distance[names(treat),, drop = FALSE]
+        if (!is.null(colnames(distance))) distance <- distance[,names(treat), drop = FALSE]
+        distance <- distance[treat == 1, treat == 0, drop = FALSE]
+      }
+      else if (all(dim.distance == c(sum(treat==1), sum(treat==0)))) {
+        if (!is.null(rownames(distance))) distance <- distance[names(treat)[treat == 1],, drop = FALSE]
+        if (!is.null(colnames(distance))) distance <- distance[,names(treat)[treat == 0], drop = FALSE]
+      }
+      else {
+        stop("When supplied as a matrix, 'distance' must have dimensions NxN or N1xN0. See ?distance for details.", call. = FALSE)
+      }
+    }
+    else {
+      if (length(distance) != length(treat)) stop("'distance' must be the same length as the dataset if specified as a numeric vector.", call. = FALSE)
+    }
+
+    if (anyNA(distance)) stop("Missing values are not allowed in 'distance'.", call. = FALSE)
   }
   return(distance)
 }
 
 #Function to check ratio is acceptable
-process.ratio <- function(ratio, min.controls = NULL, max.controls = NULL) {
-  if (length(ratio) == 0) ratio <- 1
-  if (!is.atomic(ratio) || !is.numeric(ratio) || length(ratio) > 1 || ratio < 1) {
-    stop("'ratio' must be a single positive number.", call. = FALSE)
+process.ratio <- function(ratio, method = NULL, ..., min.controls = NULL, max.controls = NULL) {
+  #Should be run after process.inputs() and ignored inputs set to NULL
+  ratio.null <- length(ratio) == 0
+  ratio.na <- !ratio.null && anyNA(ratio)
+
+  if (is.null(method)) return(1)
+  else if (method %in% c("nearest", "optimal")) {
+    if (ratio.null) ratio <- 1
+    else if (ratio.na) stop("'ratio' cannot be NA.", call. = FALSE)
+    else if (!is.atomic(ratio) || !is.numeric(ratio) || length(ratio) > 1 || ratio < 1) {
+      stop("'ratio' must be a single number greater than or equal to 1.", call. = FALSE)
+    }
+
+    if (is.null(max.controls)) {
+      ratio <- round(ratio)
+    }
+    else if (anyNA(max.controls) || !is.atomic(max.controls) || !is.numeric(max.controls) || length(max.controls) > 1) {
+      stop("'max.controls' must be a single positive number.", call. = FALSE)
+    }
+    else {
+      if (ratio <= 1) stop("'ratio' must be greater than 1 for variable ratio matching.", call. = FALSE)
+
+      max.controls <- ceiling(max.controls)
+      if (max.controls <= ratio) stop("'max.controls' must be greater than 'ratio' for variable ratio matching.", call. = FALSE)
+
+      if (is.null(min.controls)) min.controls <- 1
+      else if (anyNA(max.controls) || !is.atomic(max.controls) || !is.numeric(max.controls) || length(max.controls) > 1) {
+        stop("'max.controls' must be a single positive number.", call. = FALSE)
+      }
+      else min.controls <- floor(min.controls)
+
+      if (min.controls < 1) stop("'min.controls' cannot be less than 1 for variable ratio matching.", call. = FALSE)
+      else if (min.controls >= ratio) stop("'min.controls' must be less than 'ratio' for variable ratio matching.", call. = FALSE)
+    }
   }
-  if (is.null(max.controls)) {
+  else if (method == "full") {
+    if (is.null(max.controls)) max.controls <- Inf
+    else if ((anyNA(max.controls) || !is.atomic(max.controls) || !is.numeric(max.controls) || length(max.controls) > 1)) {
+      stop("'max.controls' must be a single positive number.", call. = FALSE)
+    }
+
+    if (is.null(min.controls)) min.controls <- 0
+    else if ((anyNA(min.controls) || !is.atomic(min.controls) || !is.numeric(min.controls) || length(min.controls) > 1)) {
+      stop("'min.controls' must be a single positive number.", call. = FALSE)
+    }
+
+    ratio <- 1 #Just to get min.controls and max.controls out
+  }
+  else if (method == "genetic") {
+    if (ratio.null) ratio <- 1
+    else if (ratio.na) stop("'ratio' cannot be NA.", call. = FALSE)
+    else if (!is.atomic(ratio) || !is.numeric(ratio) || length(ratio) > 1 || ratio < 1) {
+      stop("'ratio' must be a single number greater than or equal to 1.", call. = FALSE)
+    }
     ratio <- round(ratio)
-    return(c(ratio = ratio))
+
+    min.controls <- max.controls <- NULL
+  }
+  else if (method == "cardinality") {
+    if (ratio.null) ratio <- 1
+    else if (!ratio.na && (!is.atomic(ratio) || !is.numeric(ratio) || length(ratio) > 1 || ratio < 0)) {
+      stop("'ratio' must be a single positive number or NA.", call. = FALSE)
+    }
+
+    min.controls <- max.controls <- NULL
   }
   else {
-    if (ratio <= 1) stop("'ratio' must be greater than 1 for variable ratio matching.", call. = FALSE)
-
-    max.controls <- ceiling(max.controls)
-    if (max.controls <= ratio) stop("'max.controls' must be greater than 'ratio' for variable ratio matching.", call. = FALSE)
-
-    if (is.null(min.controls)) min.controls <- 1
-    else min.controls <- floor(min.controls)
-
-    if (min.controls < 1) stop("'min.controls' cannot be less than 1 for variable ratio matching.", call. = FALSE)
-    else if (min.controls >= ratio) stop("'min.controls' must be less than 'ratio' for variable ratio matching.", call. = FALSE)
-
-    return(c(ratio = ratio, min.controls = min.controls, max.controls = max.controls))
+    min.controls <- max.controls <- NULL
   }
+
+  if (!is.null(ratio)) {
+    attr(ratio, "min.controls") <- min.controls
+    attr(ratio, "max.controls") <- max.controls
+  }
+  return(ratio)
 }
 
 #Function to check if caliper is okay and process it
-process.caliper <- function(caliper = NULL, method, data = NULL, covs = NULL, mahcovs = NULL, distance = NULL, discarded = NULL, std.caliper = TRUE) {
+process.caliper <- function(caliper = NULL, method = NULL, data = NULL, covs = NULL, mahcovs = NULL, distance = NULL, discarded = NULL, std.caliper = TRUE) {
 
   #Check method; must be able to use a caliper
   #Check caliper names; if "" is one of them but distance = "mahal", throw error;
@@ -200,7 +299,7 @@ process.caliper <- function(caliper = NULL, method, data = NULL, covs = NULL, ma
   cal.in.covs <- setNames(names(caliper) %in% names(covs), names(caliper))
   cal.in.mahcovs <- setNames(names(caliper) %in% names(mahcovs), names(caliper))
   if (any(names(caliper) != "" & !cal.in.covs & !cal.in.data)) stop(paste0("All variables named in 'caliper' must be in 'data'. Variables not in 'data':\n\t",
-                                         paste0(names(caliper)[names(caliper) != "" & !cal.in.data & !cal.in.covs & !cal.in.mahcovs], collapse = ", ")), call. = FALSE)
+                                                                           paste0(names(caliper)[names(caliper) != "" & !cal.in.data & !cal.in.covs & !cal.in.mahcovs], collapse = ", ")), call. = FALSE)
 
   #Check std.caliper
   if (length(std.caliper) == 0 || !is.atomic(std.caliper) || !is.logical(std.caliper)) stop("'std.caliper' must be a logical (TRUE/FALSE) vector.", call. = FALSE)
@@ -247,6 +346,9 @@ process.caliper <- function(caliper = NULL, method, data = NULL, covs = NULL, ma
   if (anyNA(std.caliper)) stop("'std.caliper' cannot be NA.", call. = FALSE)
 
   if (any(std.caliper)) {
+    if ("" %in% names(std.caliper) && isTRUE(std.caliper[names(std.caliper) == ""]) && is.matrix(distance)) {
+      stop("When 'distance' is supplied as a matrix and a caliper for it is specified, 'std.caliper' must be FALSE for the distance measure.", call. = FALSE)
+    }
     caliper[std.caliper] <- caliper[std.caliper] * vapply(names(caliper)[std.caliper], function(x) {
       if (x == "") sd(distance[!discarded])
       else if (cal.in.data[x]) sd(data[[x]][!discarded])
@@ -257,13 +359,45 @@ process.caliper <- function(caliper = NULL, method, data = NULL, covs = NULL, ma
 
   #Add cal.formula
   if (any(names(caliper) != "" & !cal.in.covs[names(caliper)] & !cal.in.mahcovs[names(caliper)])) {
-      attr(caliper, "cal.formula") <- reformulate(names(caliper)[names(caliper) != "" & !cal.in.covs[names(caliper)] & !cal.in.mahcovs[names(caliper)]])
+    attr(caliper, "cal.formula") <- reformulate(names(caliper)[names(caliper) != "" & !cal.in.covs[names(caliper)] & !cal.in.mahcovs[names(caliper)]])
   }
 
   return(abs(caliper))
 
 }
 
+#Function to process replace argument
+process.replace <- function(replace, method = NULL, ..., reuse.max = NULL) {
+
+  if (is.null(method)) return(FALSE)
+
+  if (is.null(replace)) replace <- FALSE
+  else if (anyNA(replace) || length(replace) != 1 || !is.logical(replace)) {
+    stop("'replace' must be TRUE or FALSE.", call. = FALSE)
+  }
+
+  if (method %in% c("nearest")) {
+    if (is.null(reuse.max)) {
+      if (replace) reuse.max <- .Machine$integer.max
+      else reuse.max <- 1L
+    }
+    else if (length(reuse.max) == 1 && is.numeric(reuse.max) &&
+             (!is.finite(reuse.max) || reuse.max > .Machine$integer.max) &&
+             !anyNA(reuse.max)) {
+      reuse.max <- .Machine$integer.max
+    }
+    else if (abs(reuse.max - round(reuse.max)) > 1e-8 || length(reuse.max) != 1 ||
+             anyNA(reuse.max) || reuse.max < 1) {
+      stop("'reuse.max' must be a positive integer of length 1.", call. = FALSE)
+    }
+
+    replace <- reuse.max != 1L
+    attr(replace, "reuse.max") <- as.integer(reuse.max)
+  }
+
+  replace
+}
+
 #Function to ensure no subclass is devoid of both treated and control units by "scooting" units
 #from other subclasses. From WeightIt.
 subclass_scoot <- function(sub, treat, x, min.n = 1) {
@@ -366,6 +500,22 @@ check.package <- function(package.name, alternative = FALSE) {
   else return(invisible(TRUE))
 }
 
+#Create info component of matchit object
+create_info <- function(method, fn1, link, discard, replace, ratio, mahalanobis, subclass, antiexact, distance_is_matrix) {
+  info <- list(method = method,
+               distance = if (is.null(fn1)) NULL else sub("distance2", "", fn1, fixed = TRUE),
+               link = if (is.null(link)) NULL else link,
+               discard = discard,
+               replace = if (!is.null(method) && method %in% c("nearest", "genetic")) replace else NULL,
+               ratio = if (!is.null(method) && method %in% c("nearest", "optimal", "genetic")) ratio else NULL,
+               max.controls = if (!is.null(method) && method %in% c("nearest", "optimal")) attr(ratio, "max.controls") else NULL,
+               mahalanobis = mahalanobis,
+               subclass = if (!is.null(method) && method == "subclass") length(unique(subclass[!is.na(subclass)])) else NULL,
+               antiexact = antiexact,
+               distance_is_matrix = distance_is_matrix)
+  info
+}
+
 #Function to turn a method name into a phrase describing the method
 info.to.method <- function(info) {
 
@@ -379,8 +529,11 @@ info.to.method <- function(info) {
            "optimal" = "optimal pair matching",
            "full" = "optimal full matching",
            "genetic" = "genetic matching",
-           "subclass" = paste0("subclassification (", info$subclass, " subclasses)"))
-  out.list[["replace"]] <- if (!is.null(info$replace) && info$method %in% c("nearest", "optimal", "genetic")) {
+           "subclass" = paste0("subclassification (", info$subclass, " subclasses)"),
+           "cardinality" = "cardinality matching",
+           if (is.null(attr(info$method, "method"))) "an unspecified matching method"
+           else attr(info$method, "method"))
+  out.list[["replace"]] <- if (!is.null(info$replace) && info$method %in% c("nearest", "genetic")) {
     if (info$replace) "with replacement"
     else "without replacement"
   } else NULL
@@ -405,6 +558,19 @@ info.to.distance <- function(info) {
   else if (distance == "gam") {
     dist <- paste("GAM with a", link, "link")
   }
+  else if (distance == "gbm") {
+    dist <- "GBM"
+  }
+  else if (distance == "elasticnet") {
+    dist <- paste("an elastic net with a", link, "link")
+  }
+  else if (distance == "lasso") {
+    if (link == "logit") dist <- "lasso logistic regression"
+    else dist <- paste("lasso regression with a", link, "link")
+  }
+  else if (distance == "ridge") {
+    dist <- paste("ridge regression with a", link, "link")
+  }
   else if (distance == "rpart") {
     dist <- "CART"
   }
@@ -439,19 +605,19 @@ word_list <- function(word.list = NULL, and.or = c("and", "or"), is.are = FALSE,
 
   if (L == 0) {
     out <- ""
-    attr(out, "plural") = FALSE
+    attr(out, "plural") <- FALSE
   }
   else {
     word.list <- word.list[!word.list %in% c(NA_character_, "")]
     L <- length(word.list)
     if (L == 0) {
       out <- ""
-      attr(out, "plural") = FALSE
+      attr(out, "plural") <- FALSE
     }
     else if (L == 1) {
       out <- word.list
       if (is.are) out <- paste(out, "is")
-      attr(out, "plural") = FALSE
+      attr(out, "plural") <- FALSE
     }
     else {
       and.or <- match_arg(and.or)
@@ -464,7 +630,7 @@ word_list <- function(word.list = NULL, and.or = c("and", "or"), is.are = FALSE,
 
       }
       if (is.are) out <- paste(out, "are")
-      attr(out, "plural") = TRUE
+      attr(out, "plural") <- TRUE
     }
 
   }
@@ -472,10 +638,10 @@ word_list <- function(word.list = NULL, and.or = c("and", "or"), is.are = FALSE,
 }
 
 #Add quotation marks around a string.
-add_quotes <- function(x, quotes = 2) {
+add_quotes <- function(x, quotes = 2L) {
   if (!isFALSE(quotes)) {
-    if (isTRUE(quotes) || as.integer(quotes) == 2) x <- paste0("\"", x, "\"")
-    else if (as.integer(quotes) == 1) x <- paste0("\'", x, "\'")
+    if (isTRUE(quotes) || as.integer(quotes) == 2L) x <- paste0("\"", x, "\"")
+    else if (as.integer(quotes) == 1L) x <- paste0("\'", x, "\'")
     else stop("'quotes' must be boolean, 1, or 2.")
   }
   x
@@ -557,9 +723,9 @@ binarize <- function(variable, zero = NULL, one = NULL) {
 }
 
 #Make interaction vector out of matrix of covs
-exactify <- function(X, nam = NULL, sep = "|") {
+exactify <- function(X, nam = NULL, sep = "|", include_vars = FALSE) {
   if (is.null(nam)) nam <- rownames(X)
-  if (is.matrix(X)) X <- lapply(seq_len(ncol(X)), function(i) X[,i])
+  if (is.matrix(X)) X <- setNames(lapply(seq_len(ncol(X)), function(i) X[,i]), colnames(X))
   if (!is.list(X)) stop("X must be a matrix, data frame, or list.")
 
   #Ensure no ambiguity is created by sep
@@ -569,6 +735,17 @@ exactify <- function(X, nam = NULL, sep = "|") {
     sep0 <- paste0(sep0, sep)
   }
 
+  if (include_vars) {
+    for (i in seq_along(X)) {
+      if (is.character(X[[i]]) || is.factor(X[[i]])) {
+        X[[i]] <- paste0(names(X)[i], ' = "', X[[i]], '"')
+      }
+      else {
+        X[[i]] <- paste0(names(X)[i], ' = ', X[[i]])
+      }
+    }
+  }
+
   out <- do.call("paste", c(X, sep = sep0))
   if (!is.null(nam)) names(out) <- nam
   out
@@ -652,13 +829,20 @@ round_df_char <- function(df, digits, pad = "0", na_vals = "") {
 }
 
 #Generalized inverse; port of MASS::ginv()
-generalized_inverse <-function(sigma) {
+generalized_inverse <- function(sigma) {
   sigmasvd <- svd(sigma)
   pos <- sigmasvd$d > max(1e-8 * sigmasvd$d[1L], 0)
   sigma_inv <- sigmasvd$v[, pos, drop = FALSE] %*% (sigmasvd$d[pos]^-1 * t(sigmasvd$u[, pos, drop = FALSE]))
   return(sigma_inv)
 }
 
+#Choleski decomp for non-negative definite matrices
+chol2 <- function(Sinv) {
+  ch <- suppressWarnings(chol(Sinv, pivot = TRUE))
+  p <- order(attr(ch, "pivot"))
+  return(ch[,p])
+}
+
 #Get covariates (RHS) vars from formula
 get.covs.matrix <- function(formula = NULL, data = NULL) {
 
@@ -677,10 +861,21 @@ get.covs.matrix <- function(formula = NULL, data = NULL) {
 
   X <- model.matrix(formula, data = mf,
                     contrasts.arg = lapply(Filter(is.factor, mf),
-                                           contrasts, contrasts = FALSE))[,-1,drop = FALSE]
+                                           contrasts, contrasts = FALSE))
+  assign <- attr(X, "assign")[-1]
+  X <- X[,-1,drop=FALSE]
+  attr(X, "assign") <- assign
+
   return(X)
 }
 
+#Extracts and names the "assign" attribute from get.covs.matrix()
+get_assign <- function(mat) {
+  if (is.null(attr(mat, "assign"))) return(NULL)
+
+  setNames(attr(mat, "assign"), colnames(mat))
+}
+
 #Convert match.matrix (mm) using numerical indices to using char rownames
 nummm2charmm <- function(nummm, treat) {
   #Assumes nummm has rownames
@@ -714,7 +909,7 @@ mm2subclass <- function(mm, treat) {
 }
 
 #(Weighted) variance that uses special formula for binary variables
-wvar <- function(x, bin.var = NULL, w = NULL, na.rm = TRUE) {
+wvar <- function(x, bin.var = NULL, w = NULL) {
   if (is.null(w)) w <- rep(1, length(x))
   if (is.null(bin.var)) bin.var <- all(x == 0 | x == 1)
 
@@ -730,9 +925,30 @@ wvar <- function(x, bin.var = NULL, w = NULL, na.rm = TRUE) {
   }
 }
 
+#Weighted mean faster than weighted.mean()
+wm <- function(x, w = NULL, na.rm = TRUE) {
+  if (is.null(w)) {
+    if (anyNA(x)) {
+      if (!na.rm) return(NA_real_)
+      nas <- which(is.na(x))
+      x <- x[-nas]
+    }
+    return(sum(x)/length(x))
+  }
+  else {
+    if (anyNA(x) || anyNA(w)) {
+      if (!na.rm) return(NA_real_)
+      nas <- which(is.na(x) | is.na(w))
+      x <- x[-nas]
+      w <- w[-nas]
+    }
+    return(sum(x*w)/sum(w))
+  }
+}
+
 #Effective sample size
 ESS <- function(w) {
-  sum(w)^2/sum(w^2)
+  sum(abs(w))^2/sum(w^2)
 }
 
 #Compute sample sizes
@@ -740,9 +956,9 @@ nn <- function(treat, weights, discarded, s.weights) {
   if (is.null(s.weights)) s.weights <- rep(1, length(treat))
   weights <- weights * s.weights
   n <- matrix(0, ncol=2, nrow=6, dimnames = list(c("All (ESS)", "All", "Matched (ESS)","Matched", "Unmatched","Discarded"),
-                                                  c("Control", "Treated")))
+                                                 c("Control", "Treated")))
 
-  #                       Control                                    Treated
+  #                      Control                                    Treated
   n["All (ESS)",] <-     c(ESS(s.weights[treat==0]),                ESS(s.weights[treat==1]))
   n["All",] <-           c(sum(treat==0),                           sum(treat==1))
   n["Matched (ESS)",] <- c(ESS(weights[treat==0]),                  ESS(weights[treat==1]))
